"""
Turn them tables into plots
"""
import os
import six
import logging
import matplotlib
matplotlib.use('Agg')
from .utils import read_binned_df, weighting_vars # noqa
from .utils import decipher_filename, mask_rows  # noqa
from .plotting import plot_all, add_annotations # noqa


logger = logging.getLogger("fast_plotter")
logger.setLevel(logging.INFO)
HANDLER = logging.StreamHandler()
FORMATTER = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
HANDLER.setFormatter(FORMATTER)
logger.addHandler(HANDLER)


def arg_parser(args=None):
    from argparse import ArgumentParser
    parser = ArgumentParser(description=__doc__)
    parser.add_argument("tables", type=str, nargs="+",
                        help="Table files to process")
    parser.add_argument("-c", "--config", type=str,
                        help="YAML config to control common plotting options")
    parser.add_argument("-o", "--outdir", type=str, default="plots",
                        help="Output directory to save plots to")
    parser.add_argument("-e", "--extension", type=lambda x: x.split(","), default=["png"],
                        help="File extension for images")
    parser.add_argument("-w", "--weights", default=[], type=lambda x: x.split(","),
                        help="comma-separated list of weight schemes to plot things for")
    parser.add_argument("-d", "--data", default="data",
                        help="Regular expression to identify real data datasets from their name")
    parser.add_argument("-s", "--signal", default="signal",
                        help="Regular expression to identify signal MC datasets from their name")
    parser.add_argument("--dataset-col", default="dataset",
                        help="Name of column to be used to define multiple-lines for 1D plots")
    parser.add_argument("-l", "--lumi", default=None, type=float,
                        help="Scale the MC yields by this lumi")
    parser.add_argument("-y", "--yscale", default="log", choices=["log", "linear"],
                        help="Use this scale for the y-axis")

    def split_equals(arg):
        return arg.split("=")
    parser.add_argument("-v", "--variable", dest="variables", action="append", default=[], type=split_equals,
                        help="Define a variable to expand in the config file")
    parser.add_argument("--halt-errors", dest="continue_errors", default=True, action="store_false",
                        help="Stop at the first time an error occurs")
    return parser


def main(args=None):
    args = arg_parser().parse_args(args=args)
    config = getattr(args, "config", None)
    if config:
        args = process_cfg(config, args)

    if not os.path.exists(args.outdir):
        logger.info("Creating output directory " + args.outdir)
        os.makedirs(args.outdir)

    ran_ok = True
    for infile in args.tables:
        ran_ok &= process_one_file(infile, args)
    return 0 if ran_ok else 1


def process_cfg(cfg_file, args):
    import yaml
    from argparse import Namespace
    from string import Template
    with open(cfg_file, "r") as infile:
        cfg = yaml.safe_load(infile)
    # Only way to neatly allow cmd-line args to override config and handle
    # defaults seems to be:
    parser = arg_parser()
    parser.set_defaults(**cfg)
    args = parser.parse_args()
    if args.variables:

        def recursive_replace(value, replacements):
            if isinstance(value, (tuple, list)):
                return type(value)([recursive_replace(v, replacements) for v in value])
            if isinstance(value, dict):
                return {k: recursive_replace(v, replacements) for k, v in value.items()}
            if isinstance(value, six.string_types):
                return Template(value).safe_substitute(replacements)
            return value

        replacements = dict(args.variables)
        args = Namespace(**recursive_replace(vars(args), replacements))

    return args


def process_one_file(infile, args):
    logger.info("Processing: " + infile)
    df = read_binned_df(infile, dtype={args.dataset_col: str})
    weights = weighting_vars(df)
    ran_ok = True
    for weight in weights:
        if args.weights and weight not in args.weights:
            continue
        df_filtered = df.copy()
        if weight == "n":
            df_filtered["sumw"] = df_filtered.n
            df_filtered["sumw2"] = df_filtered.n
        else:
            if "n" in df.columns:
                data_rows = mask_rows(df_filtered,
                                      regex=args.data,
                                      level=args.dataset_col)
                for col in df_filtered.columns:
                    if col == "n":
                        continue
                    df_filtered.loc[data_rows, col] = df["n"][data_rows]
            df_filtered.columns = [
                n.replace(weight + ":", "") for n in df_filtered.columns]
        if hasattr(args, "value_replacements"):
            for column, replacements in args.value_replacements.items():
                if column not in df_filtered.index.names:
                    continue
                df_filtered.rename(replacements, level=column, inplace=True, axis="index")
                df_filtered = df_filtered.groupby(level=df.index.names).sum()
        plots, ok = plot_all(df_filtered, **vars(args))
        ran_ok &= ok
        dress_main_plots(plots, **vars(args))
        save_plots(infile, weight, plots, args.outdir, args.extension)
    return ran_ok


def dress_main_plots(plots, annotations=[], yscale=None, ylabel=None, legend={}, limits={}, **kwargs):
    for main_ax, _ in plots.values():
        add_annotations(annotations, main_ax)
        if yscale:
            main_ax.set_yscale(yscale)
        if ylabel:
            main_ax.set_ylabel(ylabel)
        main_ax.legend(**legend)
        main_ax.grid(True)
        main_ax.set_axisbelow(True)
        for axis, lims in limits.items():
            if isinstance(lims, (tuple, list)):
                lims = map(float, lims)
                if axis.lower() in "xy":
                    getattr(main_ax, "set_%slim" % axis)(*lims)
            elif lims.endswith("%"):
                main_ax.margins(**{axis: float(lims[:-1])})


def save_plots(infile, weight, plots, outdir, extensions):
    binning, name = decipher_filename(infile)
    kernel = "plot_" + ".".join(binning)
    kernel += "--" + ".".join(name)
    kernel += "--" + weight
    kernel = os.path.join(outdir, kernel)
    if not isinstance(extensions, (list, tuple)):
        extensions = [extensions]
    for properties, (main, ratio) in plots.items():
        insert = "-".join("%s_%s" % prop for prop in properties)
        path = kernel + "--" + insert
        for ext in extensions:
            path_ext = path + "." + ext
            logger.info("Saving plot: " + path_ext)
            plot = main.get_figure()
            plot.savefig(path_ext, dpi=200)
            matplotlib.pyplot.close(plot)


if __name__ == "__main__":
    main()



import pandas as pd


_interval_regex = r"^(?P<open>[[(])"
_interval_regex += r"(?P<low>-inf|[-+]?[0-9][.0-9]*)"
_interval_regex += r"\s*,\s*"
_interval_regex += r"(?P<high>\+?inf|[-+]?[0-9][.0-9]*)"
_interval_regex += r"(?P<close>[)\]])$"


def interval_from_string(series):
    if not pd.api.types.is_string_dtype(series):
        return series
    extracted = series.str.extract(_interval_regex)
    extracted = extracted.dropna()
    if len(extracted) != len(series):
        return series
    left_closed = extracted.open.unique()
    right_closed = extracted.close.unique()
    if len(left_closed) != 1 or len(right_closed) != 1:
        return series
    left_closed = left_closed[0] == "["
    right_closed = right_closed[0] == "]"
    if left_closed:
        if right_closed:
            closed = "both"
        else:
            closed = "left"
    else:
        if right_closed:
            closed = "right"
        else:
            closed = "neither"

    left = pd.to_numeric(extracted.low)
    right = pd.to_numeric(extracted.high)
    interval = pd.IntervalIndex.from_arrays(left=left,
                                            right=right,
                                            closed=closed)
    return interval


def convert_intervals(df, to="mid", level=[], column=[]):
    df = convert_intervals_level(df, to=to, select=level)
    df = convert_intervals_column(df, to=to, select=column)
    return df


def convert_intervals_level(df, to="mid", select=[]):
    def _convert_interval_index(index, to):
        if isinstance(index, pd.IntervalIndex):
            converted = getattr(index, to)
            converted.name = index.name
            return converted
        return index

    if not isinstance(df.index, pd.MultiIndex):
        df.index = _convert_interval_index(df.index, to)
        return df

    for name, level in zip(df.index.names, df.index.levels):
        if select and name not in select:
            continue
        out_level = _convert_interval_index(level, to)
        df.index = df.index.set_levels([out_level], [name])
    return df


def convert_intervals_column(df, to="mid", select=[]):
    for col in df.columns:
        if select and col not in select:
            continue
        if hasattr(df[col], to):
            df[col] = getattr(df[col], to)
    return df



from . import utils as utils
from . import statistics as stats
import traceback
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.colors as mc
import logging
logger = logging.getLogger(__name__)


def change_brightness(color, amount):
    if amount is None:
        return
    import colorsys
    try:
        c = mc.cnames[color]
    except Exception:
        c = color
    if isinstance(color, tuple):
        color = mc.to_rgb(c)
    c = colorsys.rgb_to_hls(*color)
    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


def plot_all(df, project_1d=True, project_2d=True, data="data", signal=None, dataset_col="dataset",
             yscale="log", lumi=None, annotations=[], dataset_order=None,
             continue_errors=True, bin_variable_replacements={}, colourmap="nipy_spectral",
             figsize=None, **kwargs):
    figures = {}

    dimensions = utils.binning_vars(df)
    ran_ok = True

    if len(dimensions) == 1:
        df = utils.rename_index(df, bin_variable_replacements)
        figures[(("yscale", yscale),)] = plot_1d(
            df, yscale=yscale, annotations=annotations)

    if dataset_col in dimensions:
        dimensions = tuple(dim for dim in dimensions if dim != dataset_col)
        if dataset_order is None:
            dataset_order = df.index.unique(dataset_col).tolist()

    if project_1d and len(dimensions) >= 1:
        for dim in dimensions:
            logger.info("Making 1D Projection: " + dim)
            projected = df.groupby(level=(dim, dataset_col)).sum()
            projected = utils.rename_index(
                projected, bin_variable_replacements)
            projected = utils.order_datasets(projected, "sum-ascending", dataset_col)
            try:
                plot = plot_1d_many(projected, data=data, signal=signal,
                                    dataset_col=dataset_col, scale_sims=lumi,
                                    colourmap=colourmap, dataset_order=dataset_order,
                                    figsize=figsize, **kwargs
                                    )
                figures[(("project", dim), ("yscale", yscale))] = plot
            except Exception as e:
                if not continue_errors:
                    raise
                logger.error("Couldn't plot 1D projection: " + dim)
                logger.error(traceback.print_exc())
                logger.error(e)
                ran_ok = False

    if project_2d and len(dimensions) > 2:
        logger.warn("project_2d is not yet implemented")

    return figures, ran_ok


class ColorDict():
    def __init__(self, order=None, named=None, n_colors=10, cmap="nipy_spectral", cmap_start=0.96, cmap_stop=0.2):
        self.order = {}
        if order is not None:
            self.order = {n: i for i, n in enumerate(order)}
            n_colors = max(n_colors, len(order))

        if isinstance(cmap, str):
            colmap_def = plt.get_cmap(cmap)
            n_colors = max(colmap_def.N, n_colors) if colmap_def.N < 256 else n_colors
        elif isinstance(cmap, dict):
            colmap_def = plt.get_cmap(cmap.get("map"))
            n_colors = cmap.get("n_colors", n_colors)
            cmap_start = cmap.get("colour_start", cmap_start)
            cmap_stop = cmap.get("colour_stop", cmap_stop)

        self.defaults = [colmap_def(i) for i in np.linspace(cmap_start, cmap_stop, n_colors)]
        self.named = named if named is not None else {}

    def get_colour(self, index=None, name=None):
        if index is None and name is None:
            raise RuntimeError("'Index' and 'name' cannot both be None")

        if name in self.named:
            return self.named[name]

        if name in self.order:
            return self.defaults[self.order[name]]

        if index is None:
            raise RuntimeError("'index' was not provided and we got an unknown named object '%s'" % name)

        return self.defaults[index]


class FillColl(object):
    def __init__(self, n_colors=10, ax=None, fill=True, line=True, dataset_colours=None,
                 colourmap="nipy_spectral", dataset_order=None, linewidth=0.5, expected_xs=None):
        self.calls = -1
        self.expected_xs = expected_xs
        self.colors = ColorDict(n_colors=n_colors, order=dataset_order,
                                named=dataset_colours, cmap=colourmap)

        self.ax = ax
        self.fill = fill
        self.line = line
        self.linewidth = linewidth

    def pre_call(self, column):
        ax = self.ax
        if not ax:
            ax = plt.gca()
        color = self.colors.get_colour(index=self.calls, name=column.name)
        x = column.index.values
        y = column.values
        return ax, x, y, color

    def __call__(self, col, **kwargs):
        ax, x, y, color = self.pre_call(col)
        if self.fill:
            draw(ax, "fill_between", x=x, ys=["y1"],
                 y1=y, label=col.name, expected_xs=self.expected_xs,
                 linewidth=0, color=color, **kwargs)
        if self.line:
            if self.fill:
                label = None
                color = "k"
                width = self.linewidth
                style = "-"
            else:
                color = None
                label = col.name
                width = 2
                style = "--"
            draw(ax, "step", x=x, ys=["y"], y=y, expected_xs=self.expected_xs,
                 color=color, linewidth=width, where="mid", label=label, linestyle=style)
        self.calls += 1


class BarColl(FillColl):
    def __call__(self, col, **kwargs):
        ax, x, y, color = self.pre_call(col)
        align = "center"
        if x.dtype.kind in 'biufc':
            align = "edge"
        facecolor = list(color)
        facecolor[-1] *= 0.5
        ax.bar(x, y, edgecolor=color, facecolor=facecolor, width=1, label=col.name, align=align)
        self.calls += 1


def actually_plot(df, x_axis, y, yerr, kind, label, ax, dataset_col="dataset",
                  dataset_colours=None, colourmap="nipy_spectral", dataset_order=None):
    expected_xs = df.index.unique(x_axis).values
    if kind == "scatter":
        draw(ax, "errorbar", x=df.reset_index()[x_axis], ys=["y", "yerr"], y=df[y], yerr=df[yerr],
             color="k", ms=3.5, fmt="o", label=label, expected_xs=expected_xs, add_ends=False)
        return
    if dataset_order is not None:
        input_datasets = df.index.unique(dataset_col)
        dataset_order = dataset_order + [d for d in input_datasets if d not in dataset_order]
    n_datasets = df.groupby(level=dataset_col).count()
    n_datasets = len(n_datasets[n_datasets != 0])

    vals = df[y].unstack(dataset_col).fillna(method="ffill", axis="columns")
    if kind == "line":
        filler = FillColl(n_datasets, ax=ax, fill=False, colourmap=colourmap,
                          dataset_colours=dataset_colours,
                          dataset_order=dataset_order, expected_xs=expected_xs)
        vals.apply(filler, axis=0, step="mid")
        return
    elif kind == "bar":
        filler = BarColl(n_datasets, ax=ax, colourmap=colourmap,
                         dataset_order=dataset_order, expected_xs=expected_xs)
        vals.apply(filler, axis=0, step="mid")
    elif kind == "fill":
        filler = FillColl(n_datasets, ax=ax, colourmap=colourmap,
                          dataset_colours=dataset_colours,
                          dataset_order=dataset_order,
                          line=False, expected_xs=expected_xs)
        vals.iloc[:, ::-1].apply(filler, axis=0, step="mid")
    elif kind == "fill-error-last":
        actually_plot(df, x_axis, y, yerr, "fill", label, ax, dataset_colours=dataset_colours,
                      dataset_col=dataset_col, colourmap=colourmap, dataset_order=dataset_order)
        summed = df.unstack(dataset_col).fillna(method="ffill", axis="columns")
        last_dataset = summed.columns.get_level_values(1)[n_datasets - 1]
        summed = summed.xs(last_dataset, level=1, axis="columns")
        x = summed.index.values
        y_down = (summed[y] - summed[yerr]).values
        y_up = (summed[y] + summed[yerr]).values
        draw(ax, "fill_between", x, ys=["y1", "y2"], y2=y_down, y1=y_up,
             color="gray", step="mid", alpha=0.7, expected_xs=expected_xs)
    else:
        raise RuntimeError("Unknown value for 'kind', '{}'".format(kind))


def standardize_values(x, y_values=[], fill_val=0, expected_xs=None, add_ends=True):
    """
    Standardize a set of arrays so they're ready to be plotted directly for matplotlib

    Algorithm:
    if any requested X values are missing:
        insert dummy values into X and Y values at the right location
    """
    if expected_xs is not None:
        x, y_values = add_missing_vals(x, expected_xs, y_values=y_values, fill_val=fill_val)

    if x.dtype.kind in 'bifc':
        x = replace_infs(x)

        if add_ends:
            x, y_values = pad_ends(x, y_values=y_values, fill_val=fill_val)
    return (x,) + tuple(y_values)


def replace_infs(x):
    """
    Replace (pos or neg) infinities at the ends of an array of floats

    Algorithm: X has +/- inf at an end, replace this X value with +/- the
    previous/next value of X +/- the mean width in X
    """
    x = x[:]  # Make a copy of the array
    is_left_inf = np.isneginf(x[0])
    is_right_inf = np.isposinf(x[-1])
    width_slice = x[1 if is_left_inf else None:-1 if is_right_inf else None]
    mean_width = width_slice[0]
    if len(width_slice) > 1:
        mean_width = np.diff(width_slice).mean()
    if is_left_inf:
        x[0] = x[1] - mean_width
    if is_right_inf:
        x[-1] = x[-2] + mean_width
    return x


def add_missing_vals(x, expected_xs, y_values=[], fill_val=0):
    """
    Check from a list of expected x values, if all occur in x.  If any are missing
    """
    insert = np.isin(expected_xs, x)
    new_ys = []
    for y in y_values:
        new = np.full_like(expected_xs, fill_val, dtype=y.dtype)
        new[insert] = y
        new_ys.append(new)
    if isinstance(expected_xs, (pd.Index, pd.MultiIndex)):
        new_x = expected_xs.values
    else:
        new_x = expected_xs.copy()
    return new_x, new_ys


def pad_ends(x, y_values=[], fill_val=0):
    """
    Insert a dummy entry to X and Y for all arrays
    """
    mean_width = x[0]
    if len(x) > 1:
        mean_width = np.diff(x).mean()

    x = np.concatenate((x[0:1] - mean_width, x, x[-1:] + mean_width), axis=0)
    new_values = [np.concatenate(([fill_val], y, [fill_val]), axis=0) for y in y_values]
    return x, tuple(new_values)


def plot_1d_many(df, prefix="", data="data", signal=None, dataset_col="dataset",
                 plot_sims="stack", plot_data="sum", plot_signal=None,
                 kind_data="scatter", kind_sims="fill-error-last", kind_signal="line",
                 scale_sims=None, summary="ratio-error-both", colourmap="nipy_spectral",
                 dataset_order=None, figsize=(5, 6), show_over_underflow=False,
                 dataset_colours=None, err_from_sumw2=False, **kwargs):
    y = "sumw"
    yvar = "sumw2"
    yerr = "err"
    if prefix:
        y = prefix + ":" + y
        yvar = prefix + ":" + yvar
        yerr = prefix + ":" + yerr

    df = utils.convert_intervals(df, to="mid")
    if not show_over_underflow:
        df = utils.drop_over_underflow(df)
    in_df_data, in_df_sims = utils.split_data_sims(
        df, data_labels=data, dataset_level=dataset_col)
    if scale_sims is not None:
        in_df_sims[y] *= scale_sims
        in_df_sims[yvar] *= scale_sims * scale_sims
    if signal:
        in_df_signal, in_df_sims = utils.split_data_sims(
            in_df_sims, data_labels=signal, dataset_level=dataset_col)
    else:
        in_df_signal = None

    if in_df_data is None or in_df_sims is None:
        summary = None
    if not summary:
        fig, main_ax = plt.subplots(1, 1, figsize=figsize)
    else:
        fig, ax = plt.subplots(
            2, 1, gridspec_kw={"height_ratios": (3, 1)}, sharex=True, figsize=figsize)
        fig.subplots_adjust(hspace=.1)
        main_ax, summary_ax = ax

    x_axis = [col for col in df.index.names if col != dataset_col]
    if len(x_axis) > 1:
        raise RuntimeError("Too many dimensions to plot things in 1D")
    if len(x_axis) == 0:
        raise RuntimeError(
            "Too few dimensions to multiple 1D graphs, use plot_1d instead")
    x_axis = x_axis[0]

    config = [(in_df_sims, plot_sims, kind_sims, "Monte Carlo", "plot_sims"),
              (in_df_data, plot_data, kind_data, "Data", "plot_data"),
              (in_df_signal, plot_signal, kind_signal, "Signal", "plot_signal"),
              ]
    for df, combine, style, label, var_name in config:
        if df is None or len(df) == 0:
            continue
        merged = _merge_datasets(df, combine, dataset_col, param_name=var_name, err_from_sumw2=err_from_sumw2)
        actually_plot(merged, x_axis=x_axis, y=y, yerr=yerr, kind=style,
                      label=label, ax=main_ax, dataset_col=dataset_col,
                      dataset_colours=dataset_colours,
                      colourmap=colourmap, dataset_order=dataset_order)
    main_ax.set_xlabel(x_axis)

    if not summary:
        return main_ax, None

    summary_ax = ax[1]
    err_msg = "Unknown value for summary, '{}'".format(summary)
    if summary.startswith("ratio"):
        main_ax.set_xlabel("")
        summed_data = _merge_datasets(
            in_df_data, "sum", dataset_col=dataset_col, err_from_sumw2=err_from_sumw2)
        summed_sims = _merge_datasets(
            in_df_sims, "sum", dataset_col=dataset_col, err_from_sumw2=err_from_sumw2)
        if summary == "ratio-error-both":
            error = "both"
        elif summary == "ratio-error-markers":
            error = "markers"
        else:
            raise RuntimeError(err_msg)
        if 'ratio_ylim' not in kwargs.keys():
            kwargs['ratio_ylim'] = [0., 2.]
        plot_ratio(summed_data, summed_sims, x=x_axis,
                   y=y, yerr=yerr, ax=summary_ax, error=error, ylim=kwargs['ratio_ylim'])
    else:
        raise RuntimeError(err_msg)
    return main_ax, summary_ax


def _merge_datasets(df, style, dataset_col, param_name="_merge_datasets", err_from_sumw2=False):
    if style == "stack":
        utils.calculate_error(df, do_rel_err=not err_from_sumw2)
        df = utils.stack_datasets(df, dataset_level=dataset_col)
    elif style == "sum":
        df = utils.sum_over_datasets(df, dataset_level=dataset_col)
        utils.calculate_error(df, do_rel_err=not err_from_sumw2)
    elif style:
        msg = "'{}' must be either 'sum', 'stack' or None. Got {}"
        raise RuntimeError(msg.format(param_name, style))
    return df


def add_annotations(annotations, ax):
    for cfg in annotations:
        cfg = cfg.copy()
        s = cfg.pop("text")
        xy = cfg.pop("position")
        cfg.setdefault("xycoords", "axes fraction")
        ax.annotate(s, xy=xy, **cfg)


def plot_1d(df, kind="line", yscale="lin"):
    fig, ax = plt.subplots(1)
    df["sumw"].plot(kind=kind)
    ax.set_axisbelow(True)
    plt.grid(True)
    plt.yscale(yscale)
    return fig


def plot_ratio(data, sims, x, y, yerr, ax, error="both", ylim=[0., 2]):
    # make sure both sides agree with the binning
    merged = data.join(sims, how="left", lsuffix="data", rsuffix="sims")
    data = merged.filter(like="data", axis="columns").fillna(0)
    data.columns = [col.replace("data", "") for col in data.columns]
    sims = merged.filter(like="sims", axis="columns")
    sims.columns = [col.replace("sims", "") for col in sims.columns]

    s, s_err = sims[y], sims[yerr]
    d, d_err = data[y], data[yerr]
    x_axis = data.reset_index()[x]

    if error == "markers":
        central, lower, upper = stats.try_root_ratio_plot(d, d_err, s, s_err)
        x_axis, central, lower, upper = standardize_values(x_axis, y_values=(central, lower, upper), add_ends=False)
        mask = (central != 0) & (lower != 0)
        ax.errorbar(x=x_axis[mask], y=central[mask], yerr=(lower[mask], upper[mask]),
                    fmt="o", markersize=4, color="k")

    elif error == "both":
        ratio = d / s
        rel_d_err = (d_err / s)
        rel_s_err = (s_err / s)

        vals = standardize_values(x_axis.values, y_values=[ratio, rel_s_err, rel_d_err], add_ends=False)
        x_axis, ratio, rel_s_err, rel_d_err = vals

        ax.errorbar(x=x_axis, y=ratio, yerr=rel_d_err, fmt="o", markersize=4, color="k")
        draw(ax, "fill_between", x_axis, ys=["y1", "y2"],
             y2=1 + rel_s_err, y1=1 - rel_s_err, fill_val=1,
             color="gray", step="mid", alpha=0.7)

    ax.set_ylim(ylim)
    ax.grid(True)
    ax.set_axisbelow(True)
    ax.set_xlabel(x)
    ax.set_ylabel("Data / MC")


def draw(ax, method, x, ys, **kwargs):
    fill_val = kwargs.pop("fill_val", 0)
    expected_xs = kwargs.pop("expected_xs", None)
    add_ends = kwargs.pop("add_ends", True)
    if x.dtype.kind in 'biufc':
        values = standardize_values(x, [kwargs[y] for y in ys],
                                    fill_val=fill_val,
                                    add_ends=add_ends,
                                    expected_xs=expected_xs)
        x = values[0]
        new_ys = values[1:]
        kwargs.update(dict(zip(ys, new_ys)))
        ticks = None
    else:
        x, ticks = np.arange(len(x)), x
    getattr(ax, method)(x=x, **kwargs)
    if ticks is not None:
        ax.set_xticks(x)
        ax.set_xticklabels(ticks)
    return x, ticks



import os
import six
import numpy as np
import pandas as pd
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class BinningDimCombiner():
    def __init__(self, index, cols_to_merge, combine_dims_ignore=None, delimiter="__"):
        col_names = index.names
        self.col_names = col_names
        self.cols_to_merge = [col_names.index(col) for col in cols_to_merge]
        self.cols_not_merged = [i for i, name in enumerate(col_names) if name not in cols_to_merge]
        self.ignore = combine_dims_ignore
        self.delimiter = delimiter

    def __call__(self, row):
        unmerged = tuple([row[i] for i in self.cols_not_merged])
        if self.ignore is not None:
            merged = [str(row[i]) for i in self.cols_to_merge if row[i] != self.ignore]
        else:
            merged = [str(row[i]) for i in self.cols_to_merge]
        result = (self.delimiter.join(merged), ) + unmerged
        return result


def handle_one_df(df, query=None, replacements=[],
                  combine_dims=[], combine_dims_ignore=None, combine_delim="__"):
    if query:
        df.query(query, inplace=True)
    if df.empty:
        return
    df.drop(df.filter(like="Unnamed").columns, axis=1, inplace=True)
    for column, repl in replacements:
        if column not in df.index.names:
            continue
        df.rename(repl, level=column, inplace=True, axis="index")
    df = df.groupby(level=df.index.names).sum()
    if combine_dims:
        names = df.index.names
        combiner = BinningDimCombiner(df.index, combine_dims,
                                      combine_dims_ignore,
                                      delimiter=combine_delim)
        new_index = df.index.to_frame().apply(combiner, axis="columns", result_type="expand")
        df.index = pd.MultiIndex.from_frame(new_index)
        df.index.names = combiner(names)
    return df


def query(df, query):
    """
    Keep only rows that satisfy requirements of the query string,
    See: https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html
    """
    logger.info("Applying query: %s", query)
    return handle_one_df(df, query=query)


def rebin(df, axis, mapping, ignore_when_combining=None, rename=None, drop_others=False):
    """
    Rename and / or collect bins and categories together
    """
    logger.info("Rebinning on axis, '%s'", axis)
    if isinstance(axis, (int, float, six.string_types)):
        replacements = [(axis, mapping)]
        out = handle_one_df(df, replacements=replacements)
        if drop_others:
            out = keep_bins(out, axis, set(mapping.values()))
        if rename is not None:
            out.index.set_names(rename, level=axis, inplace=True)
        return out

    out_df = handle_one_df(df, combine_dims=axis, combine_delim=";")

    def explode(mapping, expect_depth, prefix="", depth=0):
        exploded_map = {}
        for key, value in mapping.items():
            if isinstance(value, dict):
                result = explode(value, expect_depth, str(key) + ";", depth + 1)
                exploded_map.update(result)
            else:
                exploded_map[prefix + key] = prefix + value
        return exploded_map
    exploded_map = explode(mapping, len(axis))
    replacements = [(";".join(axis), exploded_map)]
    out_df = handle_one_df(out_df, replacements=replacements)

    if drop_others:
        out_df = keep_bins(out_df, axis, set(exploded_map.values()))

    if rename is None:
        out_df = split_dimension(out_df, axis)
    elif isinstance(rename, list) and len(rename) == len(axis):
        out_df = split_dimension(out_df, axis)
        out_df.index.set_names(rename, level=axis, inplace=True)
    else:
        out_df.index.set_names(rename, level=replacements[0], inplace=True)

    return out_df


def split_dimension(df, axis, delimeter=";"):
    """
    Split up a binning dimensions
    """
    logger.info("Splitting on axis, '%s'", axis)
    combined_name = delimeter.join(axis)
    index = df.index.to_frame()
    split_index = index[combined_name].str.split(delimeter, expand=True)
    for i, col in enumerate(axis):
        index[col] = split_index[i]
    index.drop(combined_name, axis="columns", inplace=True)
    df.set_index(pd.MultiIndex.from_frame(index), inplace=True, drop=True)
    return df


def keep_bins(df, axis, keep):
    """Keep bins on the single dimension, dropping others"""
    others = {val for val in df.index.unique(axis) if val not in keep}
    if not others:
        return df
    logger.info("Dropping values for '%s': %s", axis, str(others))
    out = df.drop(others, level=axis, axis="index")
    return out


def keep_specific_bins(df, axis, keep, expansions={}):
    """
    Keep all the specified bins and drop the others
    """
    logger.info("Keeping values based on '%s'", str(axis))
    if not isinstance(axis, list):
        out_df = keep_bins(df, axis, keep)
        keep = [k.format(**{name: v}) for name, vals in expansions.items() for v in vals for k in keep]
        return out_df

    delim = ";"
    out_df = handle_one_df(df, combine_dims=axis, combine_delim=delim)
    expanded = [(tuple(), keep)]
    for ax in axis[:-1]:
        expanded = [(last + (new, ), remaining) for last, new_items in expanded for new, remaining in new_items.items()]
    for ax in axis[-1:]:
        expanded = [keys + (new, ) for keys, vals in expanded for new in vals]
        expanded = [delim.join(vals) for vals in expanded]

    expanded = [k.format(**{name: v}) for name, vals in expansions.items() for v in vals for k in expanded]
    out_df = keep_bins(out_df, delim.join(axis), expanded)
    out_df = split_dimension(out_df, axis, delimeter=delim)
    return out_df


def combine_cols(df, format_strings):
    """Combine columns together using format strings"""
    logger.info("Combining columns based on: %s", str(format_strings))
    result_names = list(format_strings.keys())

    def apply_fmt(row):
        return [s.format(**row) for s in format_strings.values()]

    index = df.index.names
    new_df = df.reset_index()
    results = new_df.apply(apply_fmt, axis="columns", result_type="expand")
    results.columns = result_names
    new_df = new_df.assign(**results)
    new_df.set_index(index, inplace=True, drop=True)
    return new_df


def regex_split_dimension(df, axis, regex):
    """
    Split up a binning dimensions using a regex and pandas.Series.str.extract
    """
    logger.info("Splitting on axis, '%s' with regex '%s'", axis, regex)
    index = df.index.to_frame()
    split_index = index[axis].str.extract(regex, expand=True)
    index.drop(axis, axis="columns", inplace=True)
    for col in split_index.columns:
        index[col] = split_index[col]
    df.set_index(pd.MultiIndex.from_frame(index), inplace=True, drop=True)
    return df


def rename_cols(df, mapping):
    """Rename one or more value columns"""
    df.rename(mapping, inplace=True, axis="columns")
    return df


def rename_dim(df, mapping):
    """
    Rename one or more dimensions
    """
    df.index.names = [mapping.get(n, n) for n in df.index.names]
    return df


def split(df, axis, keep_split_dim, return_meta=True):
    """
    split the dataframe into a list of dataframes using a given binning
    dimensions
    """
    def to_tuple(obj):
        if isinstance(obj, (list, tuple)):
            return tuple(obj)
        else:
            return (obj, )

    axis = to_tuple(axis)
    logger.info("Splitting on axis: '%s'", axis)
    out_dfs = []
    groups = df.groupby(level=axis, group_keys=keep_split_dim)
    for split_val, group in groups:
        split_val = to_tuple(split_val)
        if not keep_split_dim:
            group.index = group.index.droplevel(axis)
        result = group.copy()
        if return_meta:
            meta = dict(zip(axis, split_val))
            split_name = "--".join(map("_".join, meta.items()))
            meta["split_name"] = split_name
            result = (result, meta)
        out_dfs.append(result)
    return out_dfs


def reorder_dimensions(df, order):
    """
    Reorder the binning dimensions
    """
    logger.info("Ordering binning dimensions according to: %s", str(order))
    return df.reorder_levels(order, axis="index")


def densify(df, known={}):
    """
    Densify axes with known values and / or values from axes on other
    dimensions
    """
    logger.info("Densify dimensions")
    before_len = len(df)

    index_vals = []
    names = df.index.names
    for dim in range(df.index.nlevels):
        vals = df.index.unique(dim)
        _known = known.get(df.index.names[dim], None)
        if _known is None:
            _known = known.get(dim, None)
        if _known is not None:
            vals = vals.union(_known)
        index_vals.append(vals)

    df = df.reindex(pd.MultiIndex.from_product(index_vals, names=names), axis="index")

    after_len = len(df)
    logger.info("    Change to number of bins (after - before): %d", after_len - before_len)
    return df


def stack_weights(df, drop_n_col=False):
    """
    Convert from a wide-form to long-form binned dataframe
    """
    logger.info("Stacking weights")
    if drop_n_col:
        df.drop("n", inplace=True, axis="columns")
    df.columns = pd.MultiIndex.from_tuples([c.split(":") for c in df.columns], names=["systematic", ""])
    out = df.stack(0, dropna=False)
    return out


def to_datacard_inputs(df, select_data, rename_syst_vars=False):
    """
    Convert to a long-form dataframe suitable as input to fast-datacard
    """
    logger.info("Converting to datacard inputs")
    if rename_syst_vars:
        df.columns = [n.replace("_up:", "_Up:").replace("_down:", "_Down:") for n in df.columns]
    df.set_index("n", append=True, inplace=True)
    df = stack_weights(df)
    df.reset_index(level="n", inplace=True)
    data_mask = df.eval(select_data)
    df["content"] = df.n
    df["content"][~data_mask] = df.sumw
    df["error"] = df.content / np.sqrt(df.n)

    df.drop(["n", "sumw", "sumw2"], inplace=True, axis="columns")
    return df


def assign_col(df, assignments={}, evals={}, drop_cols=[]):
    """
    Add or change columns by assigning or evaluating new ones
    """
    for key, value in evals.items():
        df[key] = df.eval(value)

    if assignments:
        df = df.assign(**assignments)

    if drop_cols:
        df.drop(drop_cols, axis="columns", inplace=True)
    return df


def assign_dim(df, assignments={}, evals={}, drop_cols=[]):
    """
    Add binning dimensions to the index by assigning or evaluating
    """
    index = df.index.to_frame()
    index = assign_col(index, assignments=assignments, evals=evals, drop_cols=drop_cols)
    df.set_index(pd.MultiIndex.from_frame(index), inplace=True, drop=True)
    return df


def merge(dfs):
    """ Merge a list of binned dataframes together
    """
    logger.info("Merging %d dataframes", len(dfs))
    final_df = pd.concat(dfs, sort=True)  # .fillna(float("-inf"))
    final_df = final_df.groupby(level=final_df.index.names).sum()  # .replace(float("-inf"), float("nan"))
    return final_df


def multiply_values(df, constant=0, mapping={}, weight_by_dataframes=[], apply_if=None):
    logger.info("Multiplying values")
    mask = slice(None)
    if apply_if:
        mask = df.eval(apply_if)
        if mask.dtype.kind != "b":
            msg = "'apply_if' statement doesn't return a boolean: %s"
            raise ValueError(msg % apply_if)
        ignored = df.loc[~mask]
        df = df.loc[mask]
    if mapping:
        raise NotImplementedError("'mapping' option not yet implemented")
    if weight_by_dataframes:
        for mul_df in weight_by_dataframes:
            df = multiply_dataframe(df, mul_df)
    if constant:
        df = df * constant
    if apply_if:
        df = pd.concat([df, ignored])
    return df


def multiply_dataframe(df, multiply_df, use_column=None):
    if isinstance(multiply_df, six.string_types):
        multiply_df = open_many([multiply_df], return_meta=False)[0]
    if use_column is not None:
        multiply_df = multiply_df[use_column]
    if isinstance(multiply_df, pd.Series):
        out = df.mul(multiply_df, axis=0)
    else:
        out = df * multiply_df
    return out


def normalise_group(df, groupby_dimensions, apply_if=None, use_column=None):
    logger.info("Normalising within groups defined by: %s", str(groupby_dimensions))
    norm_to = 1 / df.groupby(level=groupby_dimensions).sum()
    normed = multiply_dataframe(df, multiply_df=norm_to, use_column=use_column)
    return normed


def open_many(file_list, return_meta=True):
    """ Open a list of dataframe files
    """
    dfs = []
    for fname in file_list:
        df = pd.read_csv(fname, comment='#')
        drop_cols = [col for col in df.columns if "Unnamed" in col]
        df.drop(drop_cols, inplace=True, axis="columns")
        index_cols = [col for col in df.columns if "sumw" not in col and col != "n"]
        df.set_index(index_cols, inplace=True, drop=True)
        if return_meta:
            name = fname
            if "--" in name:
                name = name.split("--")[-1]
                name = os.path.splitext(name)[0]
            df = (df, dict(name=name, filename=fname))
        dfs.append(df)
    return dfs


def write_out(df, meta, filename="tbl_{dims}--{name}", out_dir=None, filetype="csv"):
    """ Write a dataframe to disk
    """
    meta = meta.copy()
    meta["dims"] = ".".join(df.index.names)

    complete_file = filename.format(**meta)
    if out_dir:
        complete_file = os.path.join(out_dir, complete_file)
    os.makedirs(os.path.dirname(complete_file), exist_ok=True)
    logger.info("Writing out file '%s'", complete_file)
    if not complete_file.endswith(filetype):
        complete_file += "." + filetype
    if filetype == "csv":
        df.to_csv(complete_file)
    elif filetype == "hd5":
        df.to_hdf(complete_file, key="df")
    return df



import re
import six
import os
import numpy as np
import pandas as pd
from .interval_from_str import interval_from_string, convert_intervals


__all__ = ["interval_from_string", "convert_intervals"]


def decipher_filename(filename):
    decipher = re.compile(r"tbl_(?P<binning>.*?)(--(?P<names>.*)|)\.csv")
    groups = decipher.match(os.path.basename(filename))

    binning = groups.group("binning").split(".")
    name = []
    if groups.group("names"):
        name = groups.group("names").split(".")
    return binning, name


def get_read_options(filename):
    index_cols, _ = decipher_filename(filename)
    options = dict(index_col=list(range(len(index_cols))),
                   comment="#"
                   )
    return options


def read_binned_df(filename, **kwargs):
    read_opts = get_read_options(filename)
    read_opts.update(kwargs)
    df = pd.read_csv(filename, **read_opts)
    columns = df.index.names[:]
    df.reset_index(inplace=True)
    dtype = kwargs.pop("dtype", [])
    for col in df.columns:
        if col in dtype:
            continue
        df[col] = interval_from_string(df[col])
    df.set_index(columns, inplace=True)
    return df


def binning_vars(df):
    if isinstance(df.index, pd.MultiIndex):
        return tuple(df.index.names)
    return (df.index,)


def weighting_vars(df):
    weight_vars = []
    for name in df.columns:
        weight = name.split(":")[0]
        if weight not in weight_vars:
            weight_vars.append(weight)

    return tuple(weight_vars)


def mask_rows(df, regex, level=None):
    if isinstance(regex, six.string_types):
        regex = re.compile(regex)
    index = df.index
    if level is not None:
        index = index.get_level_values(level)
    data_rows = index.map(lambda x: bool(regex.search(str(x))))
    return data_rows


def split_df(df, first_values, level=0):
    if df is None:
        return None, None
    if isinstance(first_values, six.string_types):
        regex = re.compile(first_values)
        first_values = [val for val in df.index.unique(level) if regex.match(val)]
    if not first_values:
        return None, df
    second = df.drop(first_values, level=level)
    second_values = second.index.unique(level=level)
    first = df.drop(second_values, level=level)
    if len(first) == 0:
        first = None
    if len(second) == 0:
        second = None
    return first, second


def split_data_sims(df, data_labels=["data"], dataset_level="dataset"):
    return split_df(df, first_values=data_labels, level=dataset_level)


def calculate_error(df, sumw2_label="sumw2", err_label="err", inplace=True, do_rel_err=True):
    if not inplace:
        df = df.copy()
    if do_rel_err:
        root_n = np.sqrt(df["n"])
    for column in df:
        if do_rel_err and column.endswith("sumw"):
            err_name = column.replace("sumw", err_label)
            errs = np.true_divide(df[column], root_n)
            errs[~np.isfinite(errs)] = np.nan
            df[err_name] = errs
        elif not do_rel_err and sumw2_label in column:
            err_name = column.replace(sumw2_label, err_label)
            df[err_name] = np.sqrt(df[column])
    if not inplace:
        return df


def groupby_all_but(df, by=None, level=None):
    if level is None and by is None:
        raise RuntimeError("Either 'by' or 'level' must be provided")
    if level is not None and by is not None:
        raise RuntimeError(
            "Only one of 'by' or 'level' should be provided, not both")

    args = {}
    if level:
        if not isinstance(df.index, pd.MultiIndex):
            raise RuntimeError(
                "Cannot use 'level' to groupby for non-multiindex DataFrame")
        if not isinstance(level, (tuple, list)):
            level = (level,)

        group_levels = [l for l in df.index.names if l not in level]
        args["level"] = group_levels

    if by:
        if not isinstance(by, (tuple, list)):
            by = (by,)
        group_columns = [l for l in df.columns if l not in by]
        args["by"] = group_columns

    return df.groupby(**args)


def stack_datasets(df, dataset_level="dataset"):
    groups = groupby_all_but(df, level=dataset_level)
    stacked = groups.cumsum()
    return stacked


def sum_over_datasets(df, dataset_level="dataset"):
    groups = groupby_all_but(df, level=dataset_level)
    summed = groups.sum()
    return summed


def order_datasets(df, dataset_order, dataset_level="dataset", values="sumw"):
    if isinstance(dataset_order, str) and dataset_order.startswith("sum"):
        sums = df.groupby(level=dataset_level).sum()
        if dataset_order == "sum-ascending":
            dataset_order = sums.sort_values(
                by=values, ascending=True).index.tolist()
        elif dataset_order == "sum-descending":
            dataset_order = sums.sort_values(
                by=values, ascending=False).index.tolist()
        return df.reindex(dataset_order, axis=0, level=dataset_level)
    elif isinstance(dataset_order, list):
        return df.reindex(dataset_order, axis=0, level=dataset_level)
    raise RuntimeError("Bad dataset_order value")


def rename_index(df, name_replacements):
    if not isinstance(df.index, pd.MultiIndex):
        return df
    df.index.names = [name_replacements.get(n, n) for n in df.index.names]
    return df


def drop_over_underflow(df):
    index = df.index.to_frame()
    index = index.select_dtypes(exclude=['object'])
    good_rows = np.isfinite(index).all(axis=1)
    return df.loc[good_rows]



import logging
from . import stages
from .functions import open_many
from fast_flow.v1 import read_sequence_yaml
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def make_parser():
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("-d", "--debug-dfs", default=False, action="store_true",
                        help="Print a dataframe after each step")
    parser.add_argument("--debug-dfs-query", default=None,
                        help="Provide a query to select rows from the debugged dataframe")
    parser.add_argument("--debug-rows", default=5, type=int,
                        help="Number of rows to dump from debugging dataframe")
    parser.add_argument("-p", "--post-process", default=None, required=True,
                        help="A yaml to configure the post-processing step")
    parser.add_argument("-o", "--outdir", default=".",
                        help="The name of the output directory")
    parser.add_argument("files", nargs="+",
                        help="Input dataframes that need merging together")
    return parser


def read_processing_cfg(fname, out_dir):
    sequence = read_sequence_yaml(fname, backend=stages, output_dir=out_dir)
    return sequence


def dump_debug_df(dfs, debug_dfs_query=""):
    if not debug_dfs_query:
        return dfs[0][0]

    for df in dfs:
        try:
            debug_df = df[0].query(debug_dfs_query)
            if not debug_df.empty:
                return debug_df
        except NameError:
            return None

    logger.debug("No dataframes contain rows matching the debug-dfs-query")
    return None


def main(args=None):
    args = make_parser().parse_args(args=args)
    if args.debug_dfs:
        logger.setLevel(logging.DEBUG)

    dfs = open_many(args.files)
    sequence = read_processing_cfg(args.post_process, args.outdir)

    for stage in sequence:
        logger.info("Working on %d dataframes", len(dfs))
        dfs = stage(dfs)
        if args.debug_dfs:
            debug_df = dump_debug_df(dfs, args.debug_dfs_query)
            if debug_df is not None:
                logger.debug(debug_df.head(args.debug_rows).to_string())


if __name__ == "__main__":
    main()



from . import functions


def _unique_vals(entries):
    from collections import defaultdict
    unique = defaultdict(set)
    for entry in entries:
        for key, val in entry.items():
            unique[key].add(val)
    unique = {k: tuple(v) if len(v) > 1 else v.pop() for k, v in unique.items()}
    return unique


class BaseManipulator():
    give_meta = False

    def __init__(self, **kwargs):
        use_outdir = getattr(self, "use_outdir", False)
        if not use_outdir:
            kwargs.pop("out_dir", None)
        self.name = kwargs.pop("name")
        self.kwargs = kwargs
        self.func = getattr(functions, self.func)
        self.doc = self.func.__doc__

    def __call__(self, dfs):
        if self.cardinality == "many-to-one":
            meta = _unique_vals([d[1] for d in dfs])
            dfs = [d[0] for d in dfs]
            out = [(self.func(dfs, **self.kwargs), meta)]
        elif self.cardinality == "one-to-many":
            self.kwargs.setdefault("return_meta", True)
            out = []
            for df, meta in dfs:
                results = self.func(df, **self.kwargs)
                results = [(df, _unique_vals([meta, m])) for df, m in results]
                out += results
        elif self.cardinality == "one-to-one":
            if self.give_meta:
                out = [(self.func(df, meta=m, **self.kwargs), m) for df, m in dfs]
            else:
                out = [(self.func(df, **self.kwargs), m) for df, m in dfs]
        elif self.cardinality == "none-to-many":
            self.kwargs.setdefault("return_meta", True)
            out = dfs + self.func(**self.kwargs)
        return out


class Query(BaseManipulator):
    cardinality = "one-to-one"
    func = "query"


class ReBin(BaseManipulator):
    cardinality = "one-to-one"
    func = "rebin"


class KeepSpecificBins(BaseManipulator):
    cardinality = "one-to-one"
    func = "keep_specific_bins"


class CombineColumns(BaseManipulator):
    cardinality = "one-to-one"
    func = "combine_cols"


class RegexSplitDimension(BaseManipulator):
    cardinality = "one-to-one"
    func = "regex_split_dimension"


class RenameCols(BaseManipulator):
    cardinality = "one-to-one"
    func = "rename_cols"


class RenameBinningDimension(BaseManipulator):
    cardinality = "one-to-one"
    func = "rename_dim"


class Split(BaseManipulator):
    cardinality = "one-to-many"
    func = "split"


class ReorderDimensions(BaseManipulator):
    cardinality = "one-to-one"
    func = "reorder_dimensions"


class Densify(BaseManipulator):
    cardinality = "one-to-one"
    func = "densify"


class StackWeights(BaseManipulator):
    cardinality = "one-to-one"
    func = "stack_weights"


class ToDatacardInputs(BaseManipulator):
    cardinality = "one-to-one"
    func = "to_datacard_inputs"


class AssignCol(BaseManipulator):
    cardinality = "one-to-one"
    func = "assign_col"


class AssignDim(BaseManipulator):
    cardinality = "one-to-one"
    func = "assign_dim"


class Merge(BaseManipulator):
    cardinality = "many-to-one"
    func = "merge"


class MultiplyValues(BaseManipulator):
    cardinality = "one-to-one"
    func = "multiply_values"


class NormaliseGroup(BaseManipulator):
    cardinality = "one-to-one"
    func = "normalise_group"


class OpenMany(BaseManipulator):
    cardinality = "none-to-many"
    func = "open_many"


class WriteOut(BaseManipulator):
    cardinality = "one-to-one"
    func = "write_out"
    use_outdir = True
    give_meta = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import os
from setuptools import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()


#with open('HISTORY.rst') as history_file:
#    history = history_file.read()


def get_version():
    _globals = {}
    with open(os.path.join("fast_plotter", "version.py")) as version_file:
        exec(version_file.read(), _globals)
    return _globals["__version__"]


requirements = ['matplotlib', 'pandas', 'numpy', 'scipy']

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest', 'flake8', 'pytest-cov']

setup(
    author="Ben Krikler",
    author_email='fast-hep@cern.ch',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="F.A.S.T. plotter package",
    entry_points={
        'console_scripts': [
            'fast_plotter=fast_plotter.__main__:main',
            'fast_plotter_postproc=fast_plotter.postproc.__main__:main',
        ],
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme,
    include_package_data=True,
    keywords='fast_plotter',
    name='fast_plotter',
    packages=find_packages(include=['fast_plotter*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/fast-hep/fast-plotter',
    version=get_version(),
    zip_safe=False,
)



import numpy as np
from scipy.stats import beta


@np.vectorize
def mid_p_interval(total, passed, conf=0.682689492137, is_upper=True):
    alpha = 1. - conf
    alpha_min = alpha / 2
    vmin = alpha_min if is_upper else (1. - alpha_min)
    tol = 1e-9  # tolerance
    pmin = 0
    pmax = 1
    p = 0

    # treat special case for 0<passed<1
    # do a linear interpolation of the upper limit values
    if passed > 0 and passed < 1:
        p0 = mid_p_interval(total, 0.0, is_upper)
        p1 = mid_p_interval(total, 1.0, is_upper)
        p = (p1 - p0) * passed + p0
        return p

    while abs(pmax - pmin) > tol:
        p = (pmin + pmax) / 2
        # make it work for non integer using the binomial - beta relationship
        v = 0.5 * beta.pdf(p, passed + 1., total - passed + 1) / (total + 1)
        # compute the binomial cdf at passed -1
        if passed >= 1:
            v += 1 - beta.cdf(p, passed, total - passed + 1)

        if v > vmin:
            pmin = p
        else:
            pmax = p

    return p


def ratio_values(num, num_err_sq, denom, denom_err_sq,):
    num_sum = num.sum()
    num_sum_sq = num_err_sq.sum()
    denom_sum = denom.sum()
    denom_sum_sq = denom_err_sq.sum()

    numerator = np.zeros_like(num)
    denominator = np.zeros_like(denom)
    mask = num_err_sq != 0
    numerator[mask] = num[mask] * num[mask] / num_err_sq[mask]
    mask = denom_err_sq != 0
    denominator[mask] = denom[mask] * num[mask] / denom_err_sq[mask]

    wratio = np.zeros_like(numerator)
    mask = (num > 0) & (denom > 0)
    wratio[mask] = (num_err_sq[mask] / num[mask]) / (denom_err_sq[mask] / denom[mask])
    mask = (num == 0) & (denom > 0)
    wratio[mask] = (num_sum_sq / num_sum) / (denom_err_sq[mask] / denom[mask])
    mask = (num > 0) & (denom == 0)
    wratio[mask] = (num_err_sq[mask] / num[mask]) / (denom_sum_sq / denom_sum)

    ratio = np.zeros_like(numerator)
    mask = denominator != 0
    ratio[mask] = numerator[mask] / denominator[mask]
    lower = mid_p_interval(denominator, numerator, is_upper=False)
    upper = mid_p_interval(denominator, numerator, is_upper=True)

    ratio = ratio / (1 - ratio) * wratio
    lower = lower / (1 - lower) * wratio
    upper = upper / (1 - upper) * wratio

    return ratio, lower, upper


def ratio_vals2(num, num_err_sq, denom, denom_err_sq, conf=0.682689492137):
    """
    Direct port of ROOT code from  TGraphAsymmErrors::Divide with options "midp pois"
    But it doesn't seem to work properly....
    """
    psumw = num.sum()
    psumw2 = num_err_sq.sum()
    tsumw = denom.sum()
    tsumw2 = denom_err_sq.sum()

    # Set the graph to have a number of points equal to the number of histogram
    # bins
    nbins = len(num)

    # Ok, now set the points for each bin
    # (Note: the TH1 bin content is shifted to the right by one:
    #  bin=0 is underflow, bin=nbins+1 is overflow.)

    # this keeps track of the number of points added to the graph
    npoint = 0

    # loop over all bins and fill the graph
    out_eff = np.zeros(nbins)
    out_low = np.zeros(nbins)
    out_upper = np.zeros(nbins)
    for i_point in range(nbins):

        tw = denom[i_point]
        pw = num[i_point]
        tw2 = denom_err_sq[i_point]
        pw2 = num_err_sq[i_point]

        # compute ratio on the effective entries ( p and t)
        # special case is when (pw=0, pw2=0) in this case we cannot get the bin weight.
        # we use then the overall weight of the full histogram
        if pw == 0 and pw2 == 0:
            p = 0
        else:
            p = (pw * pw) / pw2

        if tw == 0 and tw2 == 0:
            t = 0
        else:
            t = (tw * tw) / tw2

        wratio = 1
        if pw > 0 and tw > 0:
            # this is the ratio of the two bin weights ( pw/p  / t/tw )
            wratio = (pw * t) / (p * tw)
        elif pw == 0 and tw > 0:
            # case p histogram has zero  compute the weights from all the histogram
            # weight of histogram - sumw2/sumw
            wratio = (psumw2 * t) / (psumw * tw)
        elif tw == 0 and pw > 0:
            # case t histogram has zero  compute the weights from all the histogram
            # weight of histogram - sumw2/sumw
            wratio = (pw * tsumw) / (p * tsumw2)
        elif p > 0:
            wratio = pw / p  # not sure if needed

        t += p

        # when not using weights (all cases) or in case of  Poisson ratio with weights
        eff = 0
        if t:
            eff = p / t

        low = mid_p_interval(t, p, conf, False)
        upper = mid_p_interval(t, p, conf, True)

        # take the intervals in eff as intervals in the Poisson ratio
        eff = eff / (1 - eff) * wratio
        low = low / (1. - low) * wratio
        upper = upper / (1. - upper) * wratio

        # Set the point center and its errors
        if not np.isinf(eff):
            out_eff[i_point] = eff
            out_low[i_point] = low
            out_upper[i_point] = upper
            npoint += 1  # we have added a point to the graph

    # if (npoint < nbins):
    #   Warning("Divide","Number of graph points is different than histogram
    #   bins - %d points have been skipped",nbins-npoint)
    return out_eff, out_low, out_upper


def ratio_root(num, num_err_sq, denom, denom_err_sq):
    import rootpy.plotting as rp
    # rootpy seems to switch this on, flooding the terminal with debugging output
    import logging
    logging.getLogger("matplotlib.font_manager").setLevel(logging.INFO)

    # Set up the histograms
    top = rp.Hist(len(num), 0, 1)
    bottom = rp.Hist(len(num), 0, 1)
    for i, (d, n, d_err, n_err) in enumerate(zip(denom, num, denom_err_sq, num_err_sq)):
        bottom[i + 1] = (d, d_err)
        top[i + 1] = (n, n_err)

    # Do the actual division
    div = rp.Graph.divide(top, bottom, "e0 midp pois")

    # Convert this back to the array of points for the ratio plots
    ratios = np.zeros_like(num)
    low = np.zeros_like(num)
    upper = np.zeros_like(num)
    filled_indices = [top.FindBin(point.x.value) - 1 for point in div]
    ratios[filled_indices] = [point.y.value for point in div]
    low[filled_indices] = [point.y.error_low for point in div]
    upper[filled_indices] = [point.y.error_hi for point in div]
    return ratios, low, upper


def try_root_ratio_plot(*args, **kwargs):
    try:
        result = ratio_root(*args, **kwargs)
        return result
    except ImportError as e:
        if "rootpy" not in str(e):
            raise
        print("\n\tWarning: Using the broken errorbar method for ratio plots.\n\tInstall ROOT and rootpy to resolve\n")
        return ratio_vals2(*args, **kwargs)
